#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include <sys/resource.h>
#include <sys/stat.h>

#define HEAP_SIZE 20*1048576  
#define MAX_SIZE 14*1048576  
#define STACK_SIZE 2*1048576 



#define NIL 0
#define CONS 1

extern char* constructors[];


typedef enum {
  Ref, Prim, Char
} RefPrimFlag;



typedef struct {
  RefPrimFlag refprimflag;
  int refprimval;
} RefPrim;

void printRefPrim(RefPrim* rp){
  switch (rp->refprimflag ){
    case Ref:  printf("(-> %d)",rp->refprimval);break;
    case Prim: printf(" %d",rp->refprimval);break;
    case Char: printf(" %c",(char)rp->refprimval);break;
  }
}

void printStack();
void printHeap();

int callingArg;


typedef struct {
  int constrNr;
  int arity;
  RefPrim* args;
} HeapNode;

void printHeapNode(HeapNode* h){
  printf("Constr<%d %d>(",h->constrNr,h->arity);
  int i;
  for (i=0;i<h->arity;i++){
    if (i>0) printf(",");
      printRefPrim(h->args+i);
  }
  printf(")");
}

HeapNode** heap;
HeapNode** gcheap;

unsigned int hp=0;
unsigned int heapSize = 909;
unsigned int lastUsedheapSize = 909;;

int nr;
int arity;

HeapNode* newCell(unsigned int arity,int nr){
  HeapNode* newNode = (HeapNode*)malloc(sizeof(HeapNode));
  newNode->constrNr = nr;
  newNode->arity = arity;
  return newNode;
}

void gc();

HeapNode* newHeapCell(){
  if (hp>=heapSize-1){
    gc();
  }

  RefPrim* args = (RefPrim*)malloc(sizeof(RefPrim)*arity);
  HeapNode* newNode = (HeapNode*)malloc(sizeof(HeapNode));

  newNode->constrNr = nr;
  newNode->arity = arity;
  newNode->args = args;

  heap[hp] = newNode;
  hp++;

  return newNode;
}


RefPrim* stack;
RefPrim* baseptr;
unsigned int sp = 0;
unsigned int sb = 0;

void unpack(){
  int heapRef = stack[--sp].refprimval;
  int nr = heap[heapRef]->constrNr;
  int i;

  if (sp+heap[heapRef]->arity+1>=STACK_SIZE){
    printf("stack full\n");
    exit(-1);
  }

  for (i=heap[heapRef]->arity-1;i>=0;i--){
    stack[sp].refprimflag = heap[heapRef]->args[i].refprimflag;
    stack[sp].refprimval = heap[heapRef]->args[i].refprimval;
    sp++;
  }
  stack[sp].refprimflag = Prim;
  stack[sp].refprimval = heap[heapRef]->arity;
  sp++;
  stack[sp].refprimflag = Prim;
  stack[sp].refprimval = nr;
  sp++;
}


int heapRef;
void project(){
  heapRef = stack[--sp].refprimval;
  stack[sp].refprimflag = heap[heapRef]->args[callingArg].refprimflag;
  stack[sp].refprimval = heap[heapRef]->args[callingArg].refprimval;
  sp++;
}


void pushPrim(){
  if (sp+1>STACK_SIZE){
    printf("stack full\n");
    exit(-1);
  }
  stack[sp].refprimflag=Prim;
  stack[sp].refprimval=callingArg;
  sp++;
}

void pushRef(){
  if (sp+1>STACK_SIZE){
    printf("stack full\n");
    exit(-1);
  }
  stack[sp].refprimflag=Ref;
  stack[sp].refprimval=callingArg;
  sp++;
}

void pop(){
  sp--;
}

void popTops(){
  sp = sp - stack[--sp].refprimval-1;
}

void push(){
  if (sp+1>STACK_SIZE){
    printf("stack full\n");
    exit(-1);
  }
  stack[sp].refprimflag = stack[sb-callingArg].refprimflag;
  stack[sp].refprimval  = stack[sb-callingArg].refprimval;
  sp++;
}


void pushtop(){
  if (sp+1>STACK_SIZE){
    printf("stack full\n");
    exit(-1);
  }
  stack[sp].refprimflag = stack[sp-callingArg-1].refprimflag;
  stack[sp].refprimval  = stack[sp-callingArg-1].refprimval;
  sp++;
}

HeapNode** allocHeap(unsigned int size) {
  HeapNode** result = (HeapNode**) malloc(size*sizeof(HeapNode*));
  if (!result) {
    printf("could not allocate heap\n");
    exit(-1);
  }
  return result;
}

unsigned int copy(unsigned int heapIndex){
  HeapNode* oldcell = heap[heapIndex];

  if (oldcell->constrNr == -1){
    return oldcell->arity;
  }else{
    HeapNode* newcell = newCell(oldcell->arity,oldcell->constrNr);
    newcell->args=oldcell->args;
    oldcell->constrNr = -1;
    oldcell->arity = lastUsedheapSize;
    unsigned int result = lastUsedheapSize;
    gcheap[lastUsedheapSize] = newcell;
    lastUsedheapSize++;

    int ar=newcell->arity;
    int ai;
    for (ai=0;ai< ar;ai++){
      newcell->args[ai].refprimflag = oldcell->args[ai].refprimflag;
      if (oldcell->args[ai].refprimflag == Ref){
        newcell->args[ai].refprimval = copy(oldcell->args[ai].refprimval);
      }else{
        newcell->args[ai].refprimval = oldcell->args[ai].refprimval;
      }
    }

    return result;
  }
}

void gc(){
  printf("Starting gc. old heap: %d",hp-1);
  int luh3  = lastUsedheapSize*3;
  int newSize = luh3 > heapSize ? luh3 : heapSize;
  if (newSize>MAX_SIZE) newSize = MAX_SIZE;
  printf(" new size: %d\n",newSize);
  gcheap = allocHeap(newSize);
  lastUsedheapSize = 0;

  int i;
  for (i=0;i<sp;i++ ){
    if ((stack+i)->refprimflag==Ref){
      (stack+i)->refprimval = copy((stack+i)->refprimval);
    }
  }

  //alten heap löschen
  //int i;
  for (i=0;i<hp;i++){
    if (heap[i]->constrNr!=-1)free(heap[i]->args);
    free(*(heap+i));
  }
  free(heap);

  //neuen heap setzen
  heap = gcheap;
  heapSize = newSize;
  hp = lastUsedheapSize;
  printf("  Ending gc. new heap: %d\n",hp);
}

void printStack() {
  int i;
  for (i = 0; i < sp; i++) {
    printRefPrim(&stack[i]);
    printf("\n");
  }
}

void printHeap(){
  printf("\nHEAP\n");
  int i;
  for (i=0;i<hp;i++){
    printf("%d:  ",i);
    printHeapNode(heap[i]);
    printf("\n");
  }
}

void ret(){
  RefPrim result = stack[--sp];
  sb = stack[--sp].refprimval;
  sp -= callingArg;
  stack[sp++] = result;
}



void printString(unsigned int node){
  printf("\"");
  while (heap[node]->constrNr!=NIL) {
    printf("%c",heap[node]->args[1].refprimval);
    node = heap[node]->args[0].refprimval;
  }
  printf("\"");
}


void printHeapCell(unsigned int node) {
  if (    heap[node]->constrNr==CONS
      &&   heap[node]->args[1].refprimflag==Char){
    printString(node);
    return;
  }
  printf("%s(", constructors[heap[node]->constrNr]);
  int i;
  for (i = heap[node]->arity - 1; i >= 0; i--) {
    if (i < heap[node]->arity - 1) printf(", ");
    if (heap[node]->args[i].refprimflag == Prim) {
      printf("%d", heap[node]->args[i].refprimval);
    } else if (heap[node]->args[i].refprimflag == Char) {
      printf("%c", (char) heap[node]->args[i].refprimval);
    } else {
      printHeapCell(heap[node]->args[i].refprimval);
    }
  }
  printf(")");
}


void printResult() {
  if (stack[0].refprimflag == Prim)
    printf("%d", stack[0].refprimval);
  else if (stack[0].refprimflag == Char)
    printf("%c", (char) stack[0].refprimval);
  else {
    printHeapCell(stack[0].refprimval);
  }
}



void mkString(char* cs, unsigned int size) {
  arity = 0;
  nr = NIL;
  HeapNode* newNode = newHeapCell();
  stack[sp].refprimflag = Ref;
  stack[sp].refprimval = hp-1;
  sp++;

  int i;
  for (i = size - 1; i >= 0; i--) {
    arity = 2;
    nr = CONS;
    newNode = newHeapCell();

    newNode->args[0].refprimflag = Ref;
    newNode->args[0].refprimval = stack[sp-1].refprimval ;
    newNode->args[1].refprimflag = Char;
    newNode->args[1].refprimval = cs[i];
    stack[sp-1].refprimflag = Ref;
    stack[sp-1].refprimval = hp-1;
  }
}


char* getString() {
  int i = 0;
  RefPrim constr = stack[sp-1];
  while (heap[constr.refprimval]->constrNr != NIL) {
    constr = heap[constr.refprimval]->args[0];
    i++;
  }
  char* result = malloc((i + 1) * sizeof(char));
  constr = stack[sp-1];
  int j;
  for (j = 0; j < i; j++) {
    result[j] = (char) heap[constr.refprimval]->args[1].refprimval;
    constr = heap[constr.refprimval]->args[0];
  }
  result[i] = '\0';
  return result;
}


typedef unsigned int sizetype;

void readFile() {
  printf("Datei soll gelesen werden\n");

  char* fileName = getString();
  struct stat fileStat;
  printf("Datei: %s\n",fileName);

  if (stat(fileName, &fileStat) == 0) {
    // Dateigröße ermitteln
    sizetype fileSize = fileStat.st_size;
    printf("%d\n",fileSize);
    // String mit Größe erstellen
    char* content = malloc((fileSize + 1) * sizeof(char));
    if (!content) {
      printf("could not allocate String constant\n");
      exit(-1);
    }
    content[fileSize] = '\0';

    // Datei öffnen (nur Lesezugriff symbolisiert durch das "r")
    FILE* file = fopen(fileName, "r");
    printf("open file %s\n",fileName);
    if (file) {
      printf("opened file %s\n",fileName);
      int i;
      for (i = 0; !feof(file); i++) {
        char c = fgetc(file);
        if (c != EOF) {
          content[i] = c;
        }
      }

      printf("making String of content\n");
      sp--;
      mkString(content,fileSize+1);
      printf("free content\n");
      free(content);
      printf("ready\n");
    }
  } else {
    mkString("geht nicht",0);
    printf("not ready\n");
  }
}

void swap() {
  RefPrimFlag rp = stack[sp - 1].refprimflag;
  int v = stack[sp - 1].refprimval;
  stack[sp - 1].refprimflag = stack[sp - 2].refprimflag;
  stack[sp - 1].refprimval = stack[sp - 2].refprimval;
  stack[sp - 2].refprimflag = rp;
  stack[sp - 2].refprimval = v;
}


void writeFile(){
  swap();
  char* fileName = getString();
  printf("Datei: %s\n",fileName);
  sp--;
  FILE* file = fopen(fileName,"w");

  // Checken ob Datei geöffnet werden konnte und dann Fehlercode zurückgeben
  if (file==NULL){
    sp--;
    pushPrim(-1);
    return ;
  }

  HeapNode constr = *(heap[stack[sp - 1].refprimval]);
  while (constr.constrNr != NIL) {
    fprintf(file,"%c",(char) constr.args[1].refprimval);
    constr = *(heap[constr.args[0].refprimval]);
  }
  fclose(file);
  pushPrim(0);
}




void appendFile(){
  swap();
  char* fileName = getString();
  sp--;
  FILE* file = fopen(fileName,"a");

  // Checken ob Datei geöffnet werden konnte und dann Fehlercode zurückgeben
  if (file==NULL){
    sp--;
    pushPrim(-1);
    return ;
  }

  HeapNode constr = *(heap[stack[sp - 1].refprimval]);
  while (constr.constrNr != NIL) {
    fprintf(file,"%c",(char) constr.args[1].refprimval);
    constr = *(heap[constr.args[0].refprimval]);
  }
  fclose(file);
  pushPrim(0);
}




int main (int argc, char **argv)
{
  const rlim_t kStackSize = 125 * 1024 * 1024;
  struct rlimit rl;
  int result;

  result = getrlimit(RLIMIT_STACK, &rl);
  if (result == 0){
    if (rl.rlim_cur < kStackSize){
      rl.rlim_cur = kStackSize;
      result = setrlimit(RLIMIT_STACK, &rl);
      if (result != 0){
	fprintf(stderr, "setrlimit returned result = %d\n", result);
      }
    }
  }
  stack=(RefPrim*)malloc(STACK_SIZE*sizeof(RefPrim));
  baseptr=stack;

  heap = allocHeap(heapSize * sizeof(HeapNode*));
  go();
  gc();
  printResult();
  printf("\n");
//  printf("\nSTACK\n");
//  printStack();
//  printHeap();
  return 0;
}

