package name.panitz.wip.stackmachine;
import java.util.Map;
import java.util.HashMap;
import java.io.*;

public class StackMachine{
  public Instruction[] code;
  public int ic = 0;

  RefOrInt[] stack = new RefOrInt[2000];
  int stc = 0;
  int stb = 0;

  HeapNode[] heap = new HeapNode[1909];
  int hc = 0;

  public String[] constructors;

  public Map<String,Integer> labels = new HashMap<>();

  void step(){
    Instruction instruction = code[ic];
    //System.out.println(ic+" "+instruction);

    switch (instruction.instr) {
    case PushInt: stack[stc++] = new RefOrInt(instruction.a1);ic++;break;
    case PushChar: stack[stc++] = new RefOrInt((char)instruction.a1);ic++;break;
    case Pop: stc--;ic++;break;
    case PopTops: stc-=stack[--stc].val;ic++;break;
    case Mult: stack[stc-2] = new RefOrInt(stack[stc-2].val * stack[stc-1].val);stc--;ic++;break;  
    case Mod:  stack[stc-2] = new RefOrInt(stack[stc-2].val % stack[stc-1].val);stc--;ic++;break;  
    case Div:  stack[stc-2] = new RefOrInt(stack[stc-2].val / stack[stc-1].val);stc--;ic++;break;  
    case Add:  stack[stc-2] = new RefOrInt(stack[stc-2].val + stack[stc-1].val);stc--;ic++;break;  
    case Sub:  stack[stc-2] = new RefOrInt(stack[stc-2].val - stack[stc-1].val);stc--;ic++;break;  
    case Eq:   stack[stc-2] = new RefOrInt(stack[stc-2].val == stack[stc-1].val?1:0);stc--;ic++;break;  
    case Neq:  stack[stc-2] = new RefOrInt(stack[stc-2].val != stack[stc-1].val?1:0);stc--;ic++;break;  
    case Lt:   stack[stc-2] = new RefOrInt(stack[stc-2].val < stack[stc-1].val?1:0);stc--;ic++;break;  
    case Gt:   stack[stc-2] = new RefOrInt(stack[stc-2].val > stack[stc-1].val?1:0);stc--;ic++;break;  
    case Le:   stack[stc-2] = new RefOrInt(stack[stc-2].val <= stack[stc-1].val?1:0);stc--;ic++;break;  
    case Ge:   stack[stc-2] = new RefOrInt(stack[stc-2].val >= stack[stc-1].val?1:0);stc--;ic++;break;  
    case And:  stack[stc-2] = new RefOrInt((stack[stc-2].val==1 && 1==stack[stc-1].val)?1:0);stc--;ic++;break;  
    case Or:   stack[stc-2] = new RefOrInt((stack[stc-2].val==1 || 1==stack[stc-1].val)?1:0);stc--;ic++;break;  
    case Jmp:  ic = labels.get(instruction.label);break;  
    case JmpTrue: 
        if (stack[stc-1].val==1){
           ic = labels.get(instruction.label);
         }else {
           ic++;
         }
         stc--;
         break;  
    case Label:  ic++;break;  
    case GlobLabel:  ic++;break;  
    case Ret: {
      int args = instruction.a1;
      if (instruction.a2 != 42) ic = stack[stb-1].val;
      else ic++;

      stack[stb-1-args] = stack[stc-1];   
      stc = stc-2-args;
      stb = stack[stb].val;

      break;  
    }
    case Push:
      stack[stc++] = stack[stb-code[ic++].a1-1];
      break;  
    case PushTop:
      stack[stc] = stack[stc-code[ic++].a1-1];
      stc++;
      break;  
    case StartCall:
      if (instruction.a1==42)  stack[stc++] = new RefOrInt(ic+1);
      stack[stc] = new RefOrInt(stb);
      stb = stc;
      stc++;
      ic++;
      break;  
    case Call: 
      if (instruction.label.equals("readFile")){
        ic++;
	break;
      }
      stack[stc++] = new RefOrInt(ic+1);
      ic = labels.get(instruction.label);
      break;  

    case PushFun: 
      stack[stc++] = new RefOrInt(labels.get(instruction.label));
      ic++;
      break;

    case FunPointerCall: 
      int nextIc = ic+1;
      ic = stack[--stc].val;
      stack[stc++] = new RefOrInt(nextIc);
      break;

    case Pack:{
      int argsNr = stack[--stc].val;  
      RefOrInt[] args = new RefOrInt[argsNr];

      for (int i= argsNr-1;i>=0;i--){
        args[i]= stack[--stc];
      }  

      stack[stc++] = new RefOrInt(true,hc);
      heap[hc++] = new HeapNode(instruction.a1,args);
      ic++;
      break;
    }
    case PackReverse:{
      int argsNr = stack[--stc].val;  
      RefOrInt[] args = new RefOrInt[argsNr];

      for (int i= 0;i<argsNr;i++){
        args[i]= stack[--stc];
      }  

      stack[stc++] = new RefOrInt(true,hc);
      heap[hc++] = new HeapNode(instruction.a1,args);
      ic++;
      break;
    }
    case Unpack:
      int heapNr = stack[--stc].val;  
      HeapNode hp = heap[heapNr];

      for (RefOrInt arg:hp.args){
        stack[stc++] = arg;
      }
      stack[stc++] = new RefOrInt(hp.args.length);
      stack[stc++] = new RefOrInt(hp.name);
      ic++;
      break;
    }
  }

  String showHeap(){
    StringBuffer result = new StringBuffer();
    for (int i=0;i<hc;i++){
      result.append(i+": "+heap[i]+"\n");
    }
    result.append("\n");
    return result.toString();
  }

  String showStack(){
    StringBuffer result = new StringBuffer();
    for (int i=0;i<stc;i++){
      result.append(stack[i]+" ");
    }
    result.append("\n");
    result.append("PC: "+ic+" BS: "+stb+" SP:"+stc+"\n");
    result.append(code[ic]+"\n");
    return result.toString();
  }

  public void run() throws IOException{
    while (ic<code.length){
//      System.out.println(showHeap());
//      System.out.println(showStack());
      step();
    }
    System.out.println(stack[0]);
    Writer w = new PrintWriter(System.out);
    printNode(stack[0],w);
    w.write("\n");
    w.flush();
    

  }

  public void printNode(RefOrInt n,Writer w)throws IOException{
    if (!n.isRef){ w.write(n.toString());}
    else{
      w.write(constructors[heap[n.val].name]);
      w.write("(");
      boolean first = true;
      for (RefOrInt arg:heap[n.val].args){
        if (first) first = false; else w.write(", ");
        printNode(arg,w);
      }
      w.write(")");
    }
  }

  public static void main(String[] args)throws IOException{
    StackMachine interp = new StackMachine();
    interp.code = new Instruction[1];
    interp.code[0] = new Instruction(Instruction.Instr.PushInt,42);
    interp.run();
  }  

  public String toString(){
    StringBuffer result = new StringBuffer();
    int i = 0;
    for (Instruction x:code){
      result.append((i++)+" "+x+"\n");
    }
    result.append("PC:  "+ic);

    return result.toString();
  }
}

