package name.panitz.wip.stackmachine;
import java.io.*;

import name.panitz.wip.stackmachine.Instruction.Instr;
import static name.panitz.wip.tree.GenCode.*;

public class WriteX86{
  public boolean _32 = true;

  String mov(){return _32?"movl":"movq";}
  String ax(){return _32?"%eax":"%rax";}
  String bx(){return _32?"%ebx":"%rbx";}
  String cx(){return _32?"%ecx":"%rcx";}
  String dx(){return _32?"%edx":"%rdx";}
  String sp(){return _32?"%esp":"%rsp";}
  String bp(){return _32?"%ebp":"%rbp";}
  String push(){return _32?"pushl":"pushq";}
  String pop(){return _32?"popl":"popq";}
  String mul(){return _32?"mull":"mulq";}
  String add(){return _32?"addl":"addq";}
  String sub(){return _32?"subl":"subq";}
  String div(){return _32?"divl":"divq";}
  String and(){return _32?"andl":"andq";}
  String or(){return _32?"orl":"orq";}

  String cmp(){return _32?"cmpl":"cmpq";}
  String jmp(){return "jmp";}
  String jne(){return "jne";}
  String jnl(){return "jnl";}


  String[] conNames;
  public Instruction[] code;

  public WriteX86(Instruction[] code,  String[] conNames){
    this.code = code;
    this.conNames = conNames;
  }

  public void generateConstructorArray(Writer out)throws IOException{
    out.write(".globl  constructors\n");
    out.write(".section        .rodata\n");
    int i=0;
    for (String c:conNames){
       out.write(".LC"+i+":\n");
       out.write(".string \""+c+"\"\n");
       i++;
    }
    out.write(".data\n");
    out.write(".align 4\n");
    out.write(".type   constructors, @object\n");
    out.write(".size   constructors, "+(4*conNames.length)+"\n");
    out.write("constructors:\n");
    i=0;
    for (String c:conNames){
      out.write(".long   .LC"+i+"\n");
      i++;
    }
  }		


  private void genCompareExpr(Writer out,String jmp)throws Exception{
    genPop(out,"%ecx");
    genPop(out,"%edx");
    out.write("    movl    %edx, %ebx\n");
    out.write("    movl    %ecx, %eax\n");
    out.write("   "+cmp()+"   "+ax()+", "+bx()+"\n");

    String endLabel = label();
    String posLabel = label();

    out.write("   "+jmp+"   "+posLabel+"\n");
    pushInt(out,0);  
    out.write("   "+jmp()+"   "+endLabel+"\n");
    out.write("   "+posLabel+":\n");
    pushInt(out,1);  
    out.write("   "+endLabel+":\n");
  }

  public void writeAssembler(Writer out)throws Exception{
    generateConstructorArray(out);
    for (Instruction in:code){
      writeAssembler(out,in);
    }
    out.close();
  }

  public void pushInt(Writer out,int prim)throws Exception{
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%edx, %eax\n");
    out.write("    movl	$1, (%eax)\n");
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%eax, %edx\n");
    out.write("    movl	$"+prim+", %eax\n");
    out.write("    movl	%eax, 4(%edx)\n");
    out.write("    movl	sp, %eax\n");
    out.write("    addl	$1, %eax\n");
    out.write("    movl	%eax, sp\n");
  }

  public void pushChar(Writer out,int prim)throws Exception{
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%edx, %eax\n");
    out.write("    movl	$2, (%eax)\n");
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%eax, %edx\n");
    out.write("    movl	$"+prim+", %eax\n");
    out.write("    movl	%eax, 4(%edx)\n");
    out.write("    movl	sp, %eax\n");
    out.write("    addl	$1, %eax\n");
    out.write("    movl	%eax, sp\n");
  }


  private void pushInt(Writer out,String reg)throws Exception{
    out.write("    movl "+reg+", %ebx\n");
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%edx, %eax\n");
    out.write("    movl	$1, (%eax)\n");
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%eax, %edx\n");
    out.write("    movl	%ebx, %eax\n");
    out.write("    movl	%eax, 4(%edx)\n");
    out.write("    movl	sp, %eax\n");
    out.write("    addl	$1, %eax\n");
    out.write("    movl	%eax, sp\n");
  }

  private void genPushRef(Writer out,String reg)throws Exception{
    out.write("    movl "+reg+", %ebx\n");
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%edx, %eax\n");
    out.write("    movl	$0, (%eax)\n");
    out.write("    movl	stack, %eax\n");
    out.write("    movl	sp, %edx\n");
    out.write("    sall	$3, %edx\n");
    out.write("    addl	%eax, %edx\n");
    out.write("    movl	%ebx, %eax\n");
    out.write("    movl	%eax, 4(%edx)\n");
    out.write("    movl	sp, %eax\n");
    out.write("    addl	$1, %eax\n");
    out.write("    movl	%eax, sp\n");


  }

  private void genPop(Writer out)throws Exception{
    out.write("    movl    sp, %eax\n");
    out.write("    subl    $1, %eax\n");
    out.write("    movl    %eax, sp\n");
  }


  private void genPop(Writer out,String reg)throws Exception{
    genPop(out);
    out.write("    movl    sp, %eax  #save pop in reg\n");
    out.write("    sall    $3, %eax\n");
    out.write("    addl    stack, %eax\n");
    out.write("    addl    $4, %eax\n");
    out.write("    movl    (%eax), "+reg+"\n");    
  }

  public void writeAssembler(Writer out,Instruction ins)throws Exception{
    switch (ins.instr) {
    case PushInt: pushInt(out,ins.a1); break;
    case PushChar: pushChar(out,ins.a1); break;

    case Mult: 
      genPop(out,"%ecx");
      genPop(out,"%edx");
      out.write("    movl    %edx, %ebx\n");
      out.write("    movl    %ecx, %eax\n");
      out.write("    mull    %ebx\n");
      pushInt(out,"%eax");
      break;

    case Add: 
      genPop(out,"%ecx");
      genPop(out,"%edx");
      out.write("    addl    %ecx, %edx\n");
      pushInt(out,"%edx");
      break;

    case Sub: 
      genPop(out,"%ecx");
      genPop(out,"%edx");
      out.write("    subl    %ecx, %edx\n");
      pushInt(out,"%edx");
      break;

    case Div:
      genPop(out,"%edx");
      genPop(out,"%ecx");
      out.write("    movl    %edx, %ebx\n");
      out.write("    movl    %ecx, %eax\n");
      out.write("    movl $0, %edx\n");
      out.write("    idivl %ebx\n");
      pushInt(out,"%eax");
      break;

    case Mod:
      genPop(out,"%edx");
      genPop(out,"%ecx");
      out.write("    movl    %edx, %ebx\n");
      out.write("    movl    %ecx, %eax\n");
      out.write("    movl $0, %edx\n");
      out.write("    idivl %ebx\n");
      pushInt(out,"%edx");
      break;

    case Eq:
      genCompareExpr(out,"je");
      break;

    case Neq:
      genCompareExpr(out,"jne");
      break;

    case Lt:
      genCompareExpr(out,"jl");
      break;

    case Le:
      genCompareExpr(out,"jle");
      break;

    case Gt:
      genCompareExpr(out,"jg");
      break;

    case Ge:
      genCompareExpr(out,"jge");
      break;

    case And: 
      genPop(out,"%ebx");
      genPop(out,"%eax");
      out.write("   "+and()+"   "+bx()+", "+ax()+"\n");
      pushInt(out,"%eax");
      break;

    case Or: 
      genPop(out,"%ebx");
      genPop(out,"%eax");
      out.write("   "+or()+"   "+bx()+", "+ax()+"\n");
      pushInt(out,"%eax");
      break;

    case Jmp: 
      out.write("   "+jmp()+"   "+ins.label+"\n");
      break;

    case Label: 
      out.write(ins.label+":\n");
      break;

    case GlobLabel: 
      out.write(".globl "+ins.label+"\n");
      out.write(ins.label+":\n");
      break;

    case JmpTrue: 
      genPop(out,"%eax");
      out.write("   "+mov()+"   $1, "+bx()+"\n");
      out.write("   "+cmp()+"   "+ax()+", "+bx()+"\n");
      out.write("   je    "+ins.label+"\n");
      break;

    case JmpGt: 
      genPop(out,"%ebx");
      genPop(out,"%eax");
//      out.write("    movl    %edx, %ebx\n");
//      out.write("    movl    %ecx, %eax\n");
      out.write("   "+cmp()+"   "+bx()+", "+ax()+"\n");
      out.write("   jg    "+ins.label+"\n");
      break;

    case Ret: 
      out.write("  "+mov()+"    $"+ins.a1+", callingArg\n");
      out.write("   call ret\n");
      if (ins.a2!=42){
        out.write("   ret   \n");
      }
      break;  

    case Push:
      out.write("  movl $"+(ins.a1+1)+", callingArg #start push\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	sp, %edx\n");
      out.write("  sall	$3, %edx\n");
      out.write("  addl	%eax, %edx\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	sb, %ebx\n");
      out.write("  movl	callingArg, %ecx\n");
      out.write("  subl	%ecx, %ebx\n");
      out.write("  movl	%ebx, %ecx\n");
      out.write("  sall	$3, %ecx\n");
      out.write("  addl	%ecx, %eax\n");
      out.write("  movl	(%eax), %eax\n");
      out.write("  movl	%eax, (%edx)\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	sp, %edx\n");
      out.write("  sall	$3, %edx\n");
      out.write("  addl	%eax, %edx\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	sb, %ebx\n");
      out.write("  movl	callingArg, %ecx\n");
      out.write("  subl	%ecx, %ebx\n");
      out.write("  movl	%ebx, %ecx\n");
      out.write("  sall	$3, %ecx\n");
      out.write("  addl	%ecx, %eax\n");
      out.write("  movl	4(%eax), %eax\n");
      out.write("  movl	%eax, 4(%edx)\n");
      out.write("  movl	sp, %eax\n");
      out.write("  addl	$1, %eax\n");
      out.write("  movl	%eax, sp\n");
      break;  

    case StartCall: 
      out.write("   "+mov() +"	 sb,"+cx()+"\n");
      pushInt(out,"%ecx");
      out.write("   "+mov() +"	 sp,"+dx()+"\n");
      out.write("   "+mov() +"	 "+dx()+", sb\n");
      break;  

    case Call: 
      out.write("   call   "+ins.label+"\n");
      break;  

    case Pack: 
      genPop(out,"%ecx");
      out.write("    movl    %ecx, arity\n");
      out.write("    movl    $"+ins.a1+", nr\n");
      out.write("    call newHeapCell\n");
			
      out.write("    movl    hp, %ecx\n");			
      out.write("    decl    %ecx\n");			
      // ecx ist heapindex des neuen Knoten

      //den Multiplizieren wir mit zwei, weil vier wegen 4 bytes sind eine Adresse 			
       out.write("    sall    $2, %ecx\n");

      //darauf wird die startadresse des Heap addiert. Wir haben die Adresse!
      out.write("    addl    heap, %ecx\n");	

       //und gehen zu diesen neuen Heapknoten 	

      out.write("    movl    (%ecx), %ecx\n");			
      //nun also zum Array der Argumente (acht weiter. erst nr und arity)
      out.write("    addl    $8, %ecx\n");			
      out.write("    movl    (%ecx), %ecx\n");			

      for (int i=0;i<ins.a2;i++){
        genPop(out);
        //hole argument vom stack
        out.write("    movl    sp, %eax\n");
        out.write("    sall    $3, %eax\n");
        out.write("    addl    stack, %eax\n");

        out.write("    movl    (%eax), %edx\n");    

        out.write("    movl    %edx, (%ecx)\n");
	out.write("    addl    $4, %ecx\n");

        out.write("    movl    4(%eax), %edx\n");

        out.write("    movl    %edx, (%ecx)\n");
	out.write("    addl    $4, %ecx\n");
      }

      out.write("    movl    hp, %ecx\n");			
      out.write("    decl    %ecx\n");			
      genPushRef(out,"%ecx"); //die Adresse im Heap drauf
      break;  


    case PackReverse: 
      genPop(out,"%ecx");
      out.write("    movl    %ecx, arity\n");
      out.write("    movl    $"+ins.a1+", nr\n");
      out.write("    call newHeapCell\n");
			
      out.write("    movl    hp, %ecx\n");			
      out.write("    decl    %ecx\n");			
      // ecx ist heapindex des neuen Knoten

      //den Multiplizieren wir mit zwei, weil vier wegen 4 bytes sind eine Adresse 			
       out.write("    sall    $2, %ecx\n");

      //darauf wird die startadresse des Heap addiert. Wir haben die Adresse!
      out.write("    addl    heap, %ecx\n");	

       //und gehen zu diesen neuen Heapknoten 	

      out.write("    movl    (%ecx), %ecx\n");			
      //nun also zum Array der Argumente (acht weiter. erst nr und arity)
      out.write("    addl    $8, %ecx\n");			
      out.write("    movl    (%ecx), %ecx\n");			
      out.write("    addl    $"+(8*(ins.a2-1))+", %ecx\n");			

      for (int i=0;i<ins.a2;i++){
        genPop(out);
        //hole argument vom stack
        out.write("    movl    sp, %eax\n");
        out.write("    sall    $3, %eax\n");
        out.write("    addl    stack, %eax\n");

        out.write("    movl    (%eax), %edx\n");    

        out.write("    movl    %edx, (%ecx)\n");
	out.write("    addl    $4, %ecx\n");

        out.write("    movl    4(%eax), %edx\n");

        out.write("    movl    %edx, (%ecx)\n");
	out.write("    subl    $12, %ecx\n");
      }

      out.write("    movl    hp, %ecx\n");			
      out.write("    decl    %ecx\n");			
      genPushRef(out,"%ecx"); //die Adresse im Heap drauf
      break;  

    case Unpack: 
      out.write("    call unpack\n");			
      break;
      
    case Project:
      out.write("  "+mov()+"    $"+ins.a1+", callingArg\n");
      out.write("    call project\n");			
      break;

    case Pop:
      genPop(out);
      break;  

    case PopTops:
      out.write("    call popTops\n");			
      break;  

    case PushTop:
      out.write("  movl $"+(ins.a1)+", %ebx\n");
      out.write("  movl	sp, %edx\n");
      out.write("  movl	%ebx, %eax\n");
      out.write("  subl	%eax, %edx\n");
      out.write("  movl	%edx, %eax\n");
      out.write("  subl	$1, %eax\n");
      out.write("  movl	%eax, %ebx\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	sp, %edx\n");
      out.write("  sall	$3, %edx\n");
      out.write("  addl	%eax, %edx\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	%ebx, %ecx\n");
      out.write("  sall	$3, %ecx\n");
      out.write("  addl	%ecx, %eax\n");
      out.write("  movl	(%eax), %eax\n");
      out.write("  movl	%eax, (%edx)\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	sp, %edx\n");
      out.write("  sall	$3, %edx\n");
      out.write("  addl	%eax, %edx\n");
      out.write("  movl	stack, %eax\n");
      out.write("  movl	%ebx, %ecx\n");
      out.write("  sall	$3, %ecx\n");
      out.write("  addl	%ecx, %eax\n");
      out.write("  movl	4(%eax), %eax\n");
      out.write("  movl	%eax, 4(%edx)\n");
      out.write("  movl	sp, %eax\n");
      out.write("  addl	$1, %eax\n");
      out.write("  movl	%eax, sp\n");

//      out.write("  call pushtop\n");
      break;  

    case PushFun: 
      pushInt(out,"$"+ins.label);
      break;

    case FunPointerCall: 
      genPop(out,ax());
      out.write("   call   *%eax\n");

      break;

    }
  }
}

