package name.panitz.wip.tree;
import java.util.Set;
import java.util.HashSet;

public class CollectFreeVars extends AbstractVisitor<Set<Var>>{
  Set<Var> boundVars;
  Set<Var> freeVars = new HashSet<>();

  Set<Var> funNames;
  Set<Var> conNames;

  public CollectFreeVars(Set<Var> conNames,Set<Var> funNames,Set<Var> boundVars){
    this.boundVars = boundVars;
    this.funNames = funNames;
    this.conNames = conNames;
  }

  public Set<Var> visit(LetExpr let) throws Exception{
    Set<Var> newBound=new HashSet<>();
    for (Pair<String,Tree> arg:let.args){
      if (!boundVars.contains(arg.e1)) newBound.add(new Var(arg.e1));
      boundVars.add(new Var(arg.e1));
      arg.e2.welcome(this);
    }

    let.expr.welcome(this);
    boundVars.removeAll(newBound);
    return freeVars;
  }

  public Set<Var> visit(CaseExpr cas) throws Exception{
    Set<Var> newBound=new HashSet<>();
    if (!cas.isVarCase){
      for (Var arg:cas.args){
        if (!boundVars.contains(arg)) newBound.add(arg);
      }
    }else if (!boundVars.contains(new Var(cas.name))){
      newBound.add(new Var(cas.name));
    }
    boundVars.addAll(newBound);
    cas.def.welcome(this);
    boundVars.removeAll(newBound);
    return freeVars;
  }

  public Set<Var> visit(ProjectionExpr match) throws Exception{
    match.expr.welcome(this);
    return freeVars;
  }

  public Set<Var> visit(MatchExpr match) throws Exception{
    match.expr.welcome(this);
    for (CaseExpr cas:match.cases){
      cas.welcome(this);
    }
    return freeVars;
  }
  public Set<Var> visit(LambdaExpr lambda) throws Exception{
    Set<Var> newBound=new HashSet<>();
    for (String arg:lambda.args){
      if (!boundVars.contains(new Var(arg))) newBound.add(new Var(arg));
    }
    boundVars.addAll(newBound);
    lambda.def.welcome(this);
    boundVars.removeAll(newBound);
    return freeVars;
  }

  public Set<Var> visit(Var v) throws Exception{
    if(!boundVars.contains(v))
      freeVars.add(v);
    return freeVars;
  }
  public Set<Var> visit(FunCall funCall) throws Exception{
    if (!conNames.contains(funCall.name)&&!funNames.contains(funCall.name) && !boundVars.contains(funCall.name))
      freeVars.add(new Var(funCall.name));

    for (Tree arg:funCall.args)arg.welcome(this);
    return freeVars;
  }

  public Set<Var> visit(IntLit il) throws Exception{ 
    return freeVars;
  }
  public Set<Var> visit(StringLit il) throws Exception{ 
    return freeVars;
  }

  public Set<Var> visit(MultExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(DivExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(ModExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(AddExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(SubExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(EqExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(NeqExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(LeExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(GeExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(LtExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(GtExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(OrExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(AndExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
    return freeVars;
  }
  public Set<Var> visit(IfExpr ex) throws Exception{
    ex.cond.welcome(this);
    ex.alt1.welcome(this);
    ex.alt2.welcome(this);
    return freeVars;
  }

}

