package name.panitz.wip.tree;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.io.*;
import java.util.stream.Collectors;
import java.util.function.Function;

import static name.panitz.wip.tree.ExceptionFunction.tryCatch;

public class Copy extends AbstractVisitor<Tree>{

  public Tree visit(Program prog) throws Exception{
    Program p
      = new Program
          (prog.dataDefs
          ,prog.funs.stream()
                    .map(tryCatch((x)->(FunDef)x.welcome(this)))
                    .collect(Collectors.toList())
          ,prog.typs);
    return p;
  }

  public Tree visit(ProjectionExpr proj) throws Exception{
    return new ProjectionExpr(proj.expr.welcome(this),proj.name);
  }

  public Tree visit(LetExpr let) throws Exception{
    return new LetExpr
     (let.args.stream()
                 .map(tryCatch((p)->new Pair<>(p.e1,p.e2.welcome(this))))
                 .collect(Collectors.toList())             
     ,let.expr.welcome(this)
     );
  }


  public Tree visit(MatchExpr match) throws Exception{
    return new MatchExpr
     (match.expr.welcome(this)
     ,match.cases.stream()
                 .map(tryCatch((x)->(CaseExpr)x.welcome(this)))
                 .collect(Collectors.toList())             
     );
  }

  public Tree visit(CaseExpr cas) throws Exception{
    CaseExpr result =  new CaseExpr(cas.name,cas.args,cas.def.welcome(this));
    result.isVarCase = cas.isVarCase;
    return result;
  }
  public Tree visit(Var v) throws Exception{  
     return new Var(v.name);
  }

  public Tree visit(FunCall f) throws Exception{
    return new FunCall
      ( f.name
      , f.args.stream()
              .map(tryCatch((x)->x.welcome(this)))
              .collect(Collectors.toList()) 
      );
  }

  public Tree visit(FunDef f) throws Exception{
    return new FunDef(f.name, f.args, f.type, f.def.welcome(this));
  }

  public Tree visit(IntLit il) throws Exception{
    return new IntLit(il.i);
  }


  public Tree visit(StringLit il) throws Exception{
    return new StringLit(il.s);
  }

  public Tree visit(IfExpr ex) throws Exception{
    return new IfExpr
      (ex.cond.welcome(this)
      ,ex.alt1.welcome(this)
      ,ex.alt2.welcome(this));
  }

  public Tree visit(MultExpr ex) throws Exception{
    return new MultExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree visit(DivExpr ex) throws Exception{
    return new DivExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(ModExpr ex) throws Exception{
    return new ModExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(AddExpr ex) throws Exception{
    return new AddExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(SubExpr ex) throws Exception{
    return new SubExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(EqExpr ex) throws Exception{
    return new EqExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(NeqExpr ex) throws Exception{
    return new NeqExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(LeExpr ex) throws Exception{
    return new LeExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(GeExpr ex) throws Exception{
    return new GeExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(LtExpr ex) throws Exception{
    return new LtExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(GtExpr ex) throws Exception{
    return new GtExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(OrExpr ex) throws Exception{
    return new OrExpr(ex.left.welcome(this),ex.right.welcome(this));
  }
  public Tree  visit(AndExpr ex) throws Exception{
    return new AndExpr(ex.left.welcome(this),ex.right.welcome(this));
  }

  public Tree visit(LambdaExpr lambda) throws Exception{
    return new LambdaExpr(lambda.args,lambda.def.welcome(this));
  }
}

