package name.panitz.wip.tree;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;

import name.panitz.wip.stackmachine.*;

public class GenCode extends AbstractVisitor<List<Instruction>>{
  List<Instruction> result = new LinkedList<>();

  public Map<String,Integer> labels = new HashMap<>();
  Map<String,Integer> varEnv = new HashMap<>();
  Map<String,FunDef> funs = new HashMap<>();
  Map<String,ConDef> cons = new HashMap<>();
  String[] conNames;
  Map<String,Integer> conNumbers = new HashMap<>();

  int start=0;

  public StackMachine getStackMachine(){
    StackMachine res = new StackMachine();
    List<Instruction> instr= new LinkedList<>();
    instr.addAll(result);
    instr.add(new Instruction(Instruction.Instr.Call,"go"));
    res.code = instr.toArray(new Instruction[0]);
    res.labels = labels;
    res.ic = result.size();
    res.constructors = conNames;
    return res;
  }

  static int lbl = 0;
  public static String label() {
    lbl++;
    return "label"+lbl;
  }


  public List<Instruction> visit(Program prog) throws Exception{
    funs = prog.welcome(new CollectFunDefs());
    cons = prog.welcome(new CollectConDefs());
    conNames = new String[prog.dataDefs.stream().map((x)->x.cons.size()).reduce(0, (x,y)->x+y)];
    int i=0;
    for (DataDef dataDef:prog.dataDefs){
      for (ConDef con:dataDef.cons){
        conNames[i]=con.name;
        conNumbers.put(con.name,i);
        i++;
      }
    }
    for (FunDef fun:prog.funs){      
      fun.welcome(this);
    }
    start = result.size();
    return result;
  }


  public List<Instruction> visit(ProjectionExpr proj)  throws Exception{
    proj.expr.welcome(this);

    int i=0;
    
    breakLabel: for (ConDef con:cons.values()){
      i=0;
      for (VarDef arg:con.args){
        if (arg.name.equals(proj.name)){
	  i = con.args.size()-i-1;
	  break breakLabel;
        }
	i++;
      }
    }
    result.add(new Instruction(Instruction.Instr.Project,i));
    return result;
  }


  public List<Instruction> visit(MatchExpr match)  throws Exception{
    String endMatch = label();
    match.expr.welcome(this);
    result.add(new Instruction(Instruction.Instr.PushTop,0));
    result.add(new Instruction(Instruction.Instr.Unpack));
    for (CaseExpr cas:match.cases){
       if (cas.isVarCase){
         result.add(new Instruction(Instruction.Instr.Pop));
         result.add(new Instruction(Instruction.Instr.PopTops));
         cas.welcome(this);
         result.add(new Instruction(Instruction.Instr.Jmp,endMatch));
       }else{
         String endCase = label();
         result.add(new Instruction(Instruction.Instr.PushTop,0));
         result.add(new Instruction(Instruction.Instr.PushInt,conNumbers.get(cas.name)));
         result.add(new Instruction(Instruction.Instr.Neq));
         result.add(new Instruction(Instruction.Instr.JmpTrue,endCase));
         result.add(new Instruction(Instruction.Instr.Pop));
         result.add(new Instruction(Instruction.Instr.Pop));
         cas.welcome(this);
         result.add(new Instruction(Instruction.Instr.Jmp,endMatch));
         result.add(new Instruction(Instruction.Instr.Label,endCase));       
         labels.put(endCase,result.size());
      }
    }
    result.add(new Instruction(Instruction.Instr.Label,endMatch));       
    labels.put(endMatch,result.size());
    return result;
  }

  public List<Instruction> visit(CaseExpr cas)  throws Exception{
    FunCall theCase = (FunCall) cas.def;
    if (!cas.isVarCase){
      for (Tree arg:theCase.args.subList(cas.args.size()+1,theCase.args.size())){
        arg.welcome(this);
      }
    } else{
      for (Tree arg:theCase.args.subList(1,theCase.args.size())){
        arg.welcome(this);
      }
    }
    result.add(new Instruction(Instruction.Instr.Jmp,theCase.name));

    result.add(new Instruction(Instruction.Instr.Label,theCase.name+"callReturn"));
    labels.put(theCase.name+"callReturn",result.size());
    return result;
  }

/*

  public List<Instruction> visit(MatchExpr match)  throws Exception{
    String endMatch = label();
    match.expr.welcome(this);
    //nochmal drauf
    result.add(new Instruction(Instruction.Instr.PushTop,0));
    //auspacken
    result.add(new Instruction(Instruction.Instr.Unpack));
    for (CaseExpr cas:match.cases){
       if (cas.isVarCase){
         //Con Nr runter, jetzt sind argumente plus argumentanzahl drauf
         result.add(new Instruction(Instruction.Instr.Pop));
         //Argumente runter, nur noch das verpackte ist drauf 
         result.add(new Instruction(Instruction.Instr.PopTops));
         //schaue zum case und mache dort alles!
         cas.welcome(this);
         //hat gematcht also fertig mit match
         result.add(new Instruction(Instruction.Instr.Jmp,endMatch));
         // es gibt kein endCase Label, da danach kein case mehr matchen kann
       }else{
         String endCase = label();
         //nummer drauf
         result.add(new Instruction(Instruction.Instr.PushTop,0));
         //zum vergleichen mit case Nummer
         result.add(new Instruction(Instruction.Instr.PushInt,conNumbers.get(cas.name)));
         //teste ungleichheit
         result.add(new Instruction(Instruction.Instr.Neq));
         //bei ungleichen zum naechsten case
         result.add(new Instruction(Instruction.Instr.JmpTrue,endCase));
	 //Nummer runter
         result.add(new Instruction(Instruction.Instr.Pop));
         //anzahl der Args runter
         result.add(new Instruction(Instruction.Instr.Pop));
         //schaue zum case und mache dort alles!
         cas.welcome(this);
         //hat gematcht also fertig mit match
         result.add(new Instruction(Instruction.Instr.Jmp,endMatch));
         result.add(new Instruction(Instruction.Instr.Label,endCase));       
         labels.put(endCase,result.size());
      }
    }
    result.add(new Instruction(Instruction.Instr.Label,endMatch));       
    labels.put(endMatch,result.size());
    return result;
  }

  public List<Instruction> visit(CaseExpr cas)  throws Exception{
    FunCall theCase = (FunCall) cas.def;
    if (!cas.isVarCase){
      for (Tree arg:theCase.args.subList(cas.args.size()+1,theCase.args.size())){
        arg.welcome(this);
      }
    } else{
      for (Tree arg:theCase.args.subList(1,theCase.args.size())){
        arg.welcome(this);
      }
    }
    result.add(new Instruction(Instruction.Instr.Jmp,theCase.name));
    result.add(new Instruction(Instruction.Instr.Label,theCase.name+"callReturn"));
    labels.put(theCase.name+"callReturn",result.size());
    return result;
  }


  public List<Instruction> visit(LetExpr let)  throws Exception{
    for(Pair<String,Tree> arg:let.args) arg.e2.welcome(this);
    FunCall theExpr = (FunCall) let.expr;
    for (Tree arg:theExpr.args.subList(let.args.size(),theExpr.args.size())){
       arg.welcome(this);
    }
    result.add(new Instruction(Instruction.Instr.Jmp,theExpr.name));

    result.add(new Instruction(Instruction.Instr.Label,theExpr.name+"callReturn"));
    labels.put(theExpr.name+"callReturn",result.size());
    return result;
  }

*/
  public List<Instruction> visit(FunDef fun)  throws Exception{
    if(fun.isSingleCalled){
      result.add(new Instruction(Instruction.Instr.Label,fun.name));
      labels.put(fun.name,result.size());
      result.add(new Instruction(Instruction.Instr.StartCall,42));
    }else{
      result.add(new Instruction(Instruction.Instr.GlobLabel,fun.name));
      labels.put(fun.name,result.size());
      result.add(new Instruction(Instruction.Instr.StartCall));
    }
    int i = fun.args.size();
    Map<String,Integer> newVarEnv = new HashMap<>();
    for (VarDef arg:fun.args){
      newVarEnv.put(arg.name,i--);  
    }
    Map<String,Integer> oldVarEnv = varEnv;
    varEnv = newVarEnv;

    fun.def.welcome(this);

    varEnv = oldVarEnv;

    if(fun.isSingleCalled){
       result.add(new Instruction(Instruction.Instr.Ret,fun.args.size(),42));
       result.add(new Instruction(Instruction.Instr.Jmp,fun.name+"callReturn"));
    }else{
      result.add(new Instruction(Instruction.Instr.Ret,fun.args.size()));
    }
    return result;
  }

  public List<Instruction> visit(Var v) throws Exception{
    Integer vS = varEnv.get(v.name);
    if (vS != null){ 
      result.add(new Instruction(Instruction.Instr.Push,vS));
      return result;
    }
    result.add(new Instruction(Instruction.Instr.PushFun,v.name));
    return result;
  }

  public List<Instruction> visit(FunCall fun)  throws Exception{
    for (Tree arg:fun.args){
      arg.welcome(this);      
    }
    Integer vS = varEnv.get(fun.name);
    if (vS!=null){
      result.add(new Instruction(Instruction.Instr.Push,vS));
      result.add(new Instruction(Instruction.Instr.FunPointerCall));
    }else {
      Integer conNr = conNumbers.get(fun.name);
      if (conNr!=null){
        result.add(new Instruction(Instruction.Instr.PushInt,fun.args.size()));
        result.add(new Instruction(Instruction.Instr.Pack,conNr,fun.args.size()));
      }else{
        result.add(new Instruction(Instruction.Instr.Call,fun.name));
      }
    }
    return result;
  }

  public List<Instruction> visit(IntLit il) {
    result.add(new Instruction(Instruction.Instr.PushInt,il.i));
    return result;
  }

  public List<Instruction> visit(StringLit li) {
    String str = li.s;
    str=str.replace("\\\\", "\\");
    str=str.replace("\\n", "\n");
    str=str.replace("\\t", "\t");
    str=str.replace("\\\"", "\"");
    str=str.replace("\\\'", "\'");
		
    result.add(new Instruction(Instruction.Instr.PushInt,0));
    result.add(new Instruction(Instruction.Instr.Pack,0,0));
    for (int i=str.length()-1;i>=0;i--){
      result.add(new Instruction(Instruction.Instr.PushChar,str.charAt(i)));
      result.add(new Instruction(Instruction.Instr.PushInt,2));
      result.add(new Instruction(Instruction.Instr.PackReverse,1,2));
    }
    return result;
  }

  public List<Instruction> visit(IfExpr ex) throws Exception{
    if (ex.cond instanceof GtExpr){
      String thenLabel = label();
      String endIfLabel = label();
      visit((OpExpr)ex.cond);
      result.add(new Instruction(Instruction.Instr.JmpGt,thenLabel));
      ex.alt2.welcome(this);
      result.add(new Instruction(Instruction.Instr.Jmp,endIfLabel));
      labels.put(thenLabel,result.size());
      result.add(new Instruction(Instruction.Instr.Label,thenLabel));
      ex.alt1.welcome(this);
      labels.put(endIfLabel,result.size());
      result.add(new Instruction(Instruction.Instr.Label,endIfLabel));
    }else{

      String thenLabel = label();
      String endIfLabel = label();

      ex.cond.welcome(this);
      result.add(new Instruction(Instruction.Instr.JmpTrue,thenLabel));
      ex.alt2.welcome(this);
      result.add(new Instruction(Instruction.Instr.Jmp,endIfLabel));
      labels.put(thenLabel,result.size());
      result.add(new Instruction(Instruction.Instr.Label,thenLabel));
      ex.alt1.welcome(this);
      labels.put(endIfLabel,result.size());
      result.add(new Instruction(Instruction.Instr.Label,endIfLabel));
    }
    return result;
  }

  public void visit(OpExpr ex) throws Exception{
    ex.left.welcome(this);
    ex.right.welcome(this);
  }
  public List<Instruction> visit(MultExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Mult));
    return result;
  }

  public List<Instruction>  visit(DivExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Div));
    return result;
  }
  public List<Instruction>  visit(ModExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Mod));
    return result;
  }
  public List<Instruction>  visit(AddExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Add));
    return result;
  }
  public List<Instruction>  visit(SubExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Sub));
    return result;
  }
  public List<Instruction>  visit(EqExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Eq));
    return result;
  }
  public List<Instruction>  visit(NeqExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Neq));
    return result;
  }
  public List<Instruction>  visit(LeExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Le));
    return result;
  }
  public List<Instruction>  visit(GeExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Ge));
    return result;
  }
  public List<Instruction>  visit(LtExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Lt));
    return result;
  }
  public List<Instruction>  visit(GtExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Gt));
    return result;
  }
  public List<Instruction>  visit(OrExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.Or));
    return result;
  }
  public List<Instruction>  visit(AndExpr ex) throws Exception{
    visit((OpExpr) ex);
    result.add(new Instruction(Instruction.Instr.And));
    return result;
  }
}

