package name.panitz.wip.tree;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.io.*;
import java.util.stream.Collector;
import java.util.stream.Collectors;

import static name.panitz.wip.tree.ExceptionFunction.tryCatch;

public class GlobalizeCases extends Copy{
  List<FunDef> caseFuns = new LinkedList<>();
  Set<Var> funNames=new HashSet<>();
  Set<Var> conNames=new HashSet<>();

  public Tree visit(Program prog) throws Exception{
    for (FunDef fun:prog.funs) funNames.add(new Var(fun.name));
    for (DataDef con:prog.dataDefs) conNames.add(new Var(con.name));

    Program p = (Program)super.visit(prog);
    p.funs.addAll(caseFuns);
    return p;
  }

  public Tree visit(CaseExpr cas) throws Exception{
    Set<Var> boundVars = new HashSet<>();
    if (cas.isVarCase)boundVars.add(new Var(cas.name)); 
    else boundVars.addAll(cas.args);
    Set<Var> freeVars = cas.def.welcome(new CollectFreeVars(conNames,funNames,boundVars));
    String caseName = caseFun(); 
    
    List<VarDef> funArgs = new LinkedList<>();

    if (cas.isVarCase){
      funArgs.add(new VarDef(cas.name));
    }else{
      funArgs.add(new VarDef("this"));
      funArgs.addAll(cas.args.stream().map(x->new VarDef(x.name)).collect(Collectors.toList()));
    }
    funArgs.addAll(freeVars.stream().map(x->new VarDef(x.name)).collect(Collectors.toList()));

    FunDef f = new FunDef(caseName,funArgs,null,cas.def.welcome(this));
    f.isSingleCalled = true;

    caseFuns.add(f);
    funNames.add(new Var(caseName));
    CaseExpr result = new CaseExpr
     (cas.name,cas.args
     ,new FunCall
       (caseName
       ,funArgs.stream()
               .map(x->new Var(x.name)).collect(Collectors.toList())));

    result.isVarCase = cas.isVarCase;
    return result;
  }

  public Tree visit(LetExpr let) throws Exception{
    Set<Var> boundVars = new HashSet<>();
    for (Pair<String,Tree> arg:let.args) boundVars.add(new Var(arg.e1));

    Set<Var> freeVars = let.expr.welcome(new CollectFreeVars(conNames,funNames,boundVars));

    String caseName = "let"+caseFun(); 
    
    List<Var> funArgs = new LinkedList<>();

    for (Pair<String,Tree> arg:let.args)
    	funArgs.add(new Var(arg.e1));

    funArgs.addAll(freeVars);

    FunDef f = new FunDef(caseName,funArgs.stream()
            .map(x->new VarDef(x.name)).collect(Collectors.toList()),null,let.expr.welcome(this));
    f.isSingleCalled = true;

    caseFuns.add(f);
    funNames.add(new Var(caseName));
    LetExpr result = new LetExpr
     (let.args.stream()
              .map(tryCatch(p->new Pair<>(p.e1,p.e2.welcome(this)))).collect(Collectors.toList())
     ,new FunCall
       (caseName
       ,funArgs.stream()
               .map(x->new Var(x.name)).collect(Collectors.toList())));

    return result;
  }


  int nr = 0;
  synchronized String caseFun(){
    return "case"+(nr++);
  }
}

