package name.panitz.wip.tree;

import java.util.List;
import java.util.LinkedList;
import java.io.*;

public class PP extends AbstractVisitor<Integer>{
  int indent = 0;
  Writer w;
  public PP(Writer w){
    this.w = w;
  }

  public Integer visit(LetExpr let) throws Exception{
    w.write("let ");
    indent += 2;
    for (Pair<String,Tree> arg:let.args){
      newLine();
      w.write(arg.e1);
      w.write(" = ");
      arg.e2.welcome(this);
    }
    indent -= 2;
    newLine();
    let.expr.welcome(this);
    return indent;
  }

  public Integer visit(CaseExpr cas) throws Exception{
    newLine();
    w.write("case ");
    w.write(cas.name);
    if (!cas.isVarCase){
      w.write("(");
      boolean first = true;
      for (Var arg:cas.args){
        if (first){first = false;}else{w.write(", ");}
        w.write(arg.name);
      }
      w.write(")");
    }
    w.write(" -> ");
    return cas.def.welcome(this);
  }


  public Integer visit(ProjectionExpr proj) throws Exception{
    proj.expr.welcome(this);
    w.write(".");
    w.write(proj.name);
    return indent;
  }


  public Integer visit(MatchExpr match) throws Exception{
    w.write("match ");
    match.expr.welcome(this);
    w.write("[");
    indent+=2;


    for (CaseExpr cas:match.cases){
      cas.welcome(this);
    }
    indent-=2;
    newLine();
    w.write("]");
    return indent;
  }


  public Integer visit(LambdaExpr lambda) throws Exception{
    w.write("\\(");
    boolean first = true;
    for (String arg:lambda.args){
      if (first){first = false;}else{w.write(", ");}
      w.write(arg);
    }
    w.write(") -> ");
    return lambda.def.welcome(this);
  }

  public Integer visit(Program prog) throws Exception{
    for (FunDef fun:prog.funs){
      fun.welcome(this);
    }

    return indent;
  }
  public Integer visit(FunDef fun) throws Exception{
    w.write("fun ");
    w.write(fun.name);
    w.write("(");
    boolean first = true;
    for (VarDef arg:fun.args){
      if (first){first=false;}else{w.write(", ");}
      w.write(arg.name);
      w.write(":");
      w.write(arg.type.toString());
   }
    w.write(") = ");
    indent+=2;
    newLine();
    fun.def.welcome(this);
    indent-=2;
    newLine();
    w.write(".");    
    newLine();

    return indent;
  }

  public Integer visit(FunCall fun) throws Exception{
    w.write(fun.name);
    w.write("(");
    boolean first = true;
    for (Tree arg:fun.args){
      if (first){first=false;}else{w.write(", ");}
      arg.welcome(this);
    }
    w.write(")");
    return indent;
  }

  public Integer visit(Var v) throws Exception{
    w.write(v.name);
    return indent;
  }


  public Integer visit(IntLit il) throws Exception{
    w.write(il.i+"");
    return indent;
  }

  public Integer visit(StringLit il) throws Exception{
    w.write("\""+il.s+"\"");
    return indent;
  }

  public Integer visit(OpExpr ex) throws Exception{
    w.write("(");
    ex.left.welcome(this);
    w.write(" "+ex.opName+" ");
    ex.right.welcome(this);
    w.write(")");
    return indent;
  }
  public Integer visit(MultExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(DivExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(ModExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(AddExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(SubExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(EqExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(NeqExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(LeExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(GeExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(LtExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(GtExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(OrExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(AndExpr ex) throws Exception{
    return visit((OpExpr)ex);
  }
  public Integer visit(IfExpr ex) throws Exception{
    w.write("if");
    indent+=2;
    newLine();
    ex.cond.welcome(this);
    indent-=2;
    newLine();
    w.write("then");
    indent+=2;
    newLine();
    ex.alt1.welcome(this);
    indent-=2;
    newLine();
    w.write("else");
    indent+=2;
    newLine();
    ex.alt2.welcome(this);
    indent-=2;
    return indent;
  }
  private void newLine() throws Exception{
    w.write("\n");
    for (int i=0;i<indent;i++) w.write(" ");
  }
}

