package name.panitz.wip.types;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import name.panitz.wip.TypeDef;
import name.panitz.wip.tree.AddExpr;
import name.panitz.wip.tree.AndExpr;
import name.panitz.wip.tree.BoolLit;
import name.panitz.wip.tree.CaseExpr;
import name.panitz.wip.tree.ConDef;
import name.panitz.wip.tree.DataDef;
import name.panitz.wip.tree.DivExpr;
import name.panitz.wip.tree.EqExpr;
import name.panitz.wip.tree.FunCall;
import name.panitz.wip.tree.FunDef;
import name.panitz.wip.tree.GeExpr;
import name.panitz.wip.tree.GtExpr;
import name.panitz.wip.tree.IfExpr;
import name.panitz.wip.tree.IntLit;
import name.panitz.wip.tree.LambdaExpr;
import name.panitz.wip.tree.LeExpr;
import name.panitz.wip.tree.LetExpr;
import name.panitz.wip.tree.LtExpr;
import name.panitz.wip.tree.MatchExpr;
import name.panitz.wip.tree.ModExpr;
import name.panitz.wip.tree.MultExpr;
import name.panitz.wip.tree.NeqExpr;
import name.panitz.wip.tree.OpExpr;
import name.panitz.wip.tree.OrExpr;
import name.panitz.wip.tree.Pair;
import name.panitz.wip.tree.Program;
import name.panitz.wip.tree.ProjectionExpr;
import name.panitz.wip.tree.StringLit;
import name.panitz.wip.tree.SubExpr;
import name.panitz.wip.tree.Tree;
import name.panitz.wip.tree.Var;
import name.panitz.wip.tree.VarDef;
import name.panitz.wip.tree.Visitor;

public class TypeChecker implements Visitor<Type>{
	Map<String,Type> env = new HashMap<String, Type>();
	
	static int varNr = 0;
	static String newName(){return "v"+(varNr++);}
	
	private static final PrimType INT_TYPE = new PrimType("int");
	private static final PrimType BOOLEAN_TYPE = new PrimType("boolean");
	private static final PrimType CHAR_TYPE = new PrimType("char");
	Map<String,FunDef> funs = new HashMap<>();
	Map<String,ConDef> cons = new HashMap<String, ConDef>();
	Map<String,TypeDef> typs = new HashMap<>();

	Substitution findSub(String funName,Type formal,Type actual) throws Exception{
		return findSub(funName, formal, actual,false);
	}
	
	Substitution findSub(String funName,Type formal,Type actual,boolean unifier) throws Exception{
		List<Type> frmls = new ArrayList<Type>();	
		frmls.add(formal);
		List<Type> acts = new ArrayList<Type>();
		acts.add(actual);
		return findSub(funName, frmls, acts,unifier);
	}

	Substitution findSub(String funName,List<Type> formal,List<Type> actual) throws Exception{
		return findSub(funName, formal, actual, false);
	}

	Substitution findMGU(String funName,Type formal,Type actual) throws Exception{
		List<Type> frmls = new ArrayList<Type>();	
		frmls.add(formal);
		List<Type> acts = new ArrayList<Type>();
		acts.add(actual);
		return findMGU(funName, frmls, acts);
	}
	
	Substitution findMGU(String funName,List<Type> formal,List<Type> actual) throws Exception{
		return findSub(funName, formal, actual, true);
	}
	
	Substitution findSub(String funName,List<Type> formal,List<Type> actual,boolean unifier) throws Exception{
		if (formal.size()!=actual.size()){
			throw new Exception("wrong number of arguments. expected: "+formal.size()+" found: "+actual.size());
		}
		Substitution result = new Substitution();
		Iterator<Type> itAct = actual.iterator();
		for (Type frm:formal){
			Type act = itAct.next();
			frm = frm.substitute(result);
			act = act.substitute(result);
			if (frm.equals(act)) continue;
			if (frm instanceof VarType){
				VarType vT = (VarType)frm;
				if(!vT.fixed){ //where is this needed?
					result.substitute(vT.name, act);
					continue;
				}
			}
			if (unifier && act instanceof VarType){
				VarType vT = (VarType)act;
				if(!vT.fixed){
					result.substitute(vT.name, frm);
					continue;
				}
			}else if(act instanceof VarType){
				throw new Exception("wrong type... expected:"+frm+" found: "+act);
			}else if (frm instanceof ADT && act instanceof ADT){
				ADT frmADT = (ADT)frm;
				ADT actADT = (ADT)act;
				if (!frmADT.name.equals(actADT.name)){
					TypeDef frmTD = typs.get(frmADT.name);
					TypeDef actTD = typs.get(actADT.name);
					if (frmTD!=null||actTD!=null){
						if (frmTD!=null){
							frm = frmTD.instantiate(frmADT.args);
						}
						if (actTD!=null){
							act = actTD.instantiate(actADT.args);
						}
						Substitution sub = findSub(funName, frm, act,unifier);
						result.substitute(sub);	
						continue;
					}else{
						throw new Exception("Clash! "+frmADT.name+" <-> "+actADT.name);					
					}
				}
				if (frmADT.args.size()!=actADT.args.size()){
					throw new Exception("Wrong number of Type parameters for type "+actADT.name+": "+frmADT.args+" <-> "+actADT.args);										
				}
				Substitution sub = findSub(funName, frmADT.args, actADT.args,unifier);
				result.substitute(sub);
				//continue;
			}else if (frm instanceof FunType && act instanceof FunType){
				Substitution sub = findSub(funName, ((FunType)frm).args, ((FunType)act).args,unifier);
				result.substitute(sub);
				result.substitute(findSub(funName, ((FunType)frm).result.substitute(result), ((FunType)act).result.substitute(result),unifier));
				//continue;
			}else if (isTypeDef(frm)||isTypeDef(act)){
				result.substitute(findSub(funName, frm.getDefinition(typs), act.getDefinition(typs), unifier));
			}else{
				throw new Exception("Typefehler: "+frm+" "+act);
			}
		}
		
		return result;
	}
	
	private boolean isTypeDef(Type frm) {
		return (frm instanceof ADT) && typs.get(((ADT)frm).name)!=null;
	}

	@Override
	public Type visit(Program prog) throws Exception {
		for (FunDef fun: prog.funs)
			funs.put(fun.name,fun);
		for (TypeDef typ: prog.typs)
			typs.put(typ.name,typ);
		for (DataDef dataDef: prog.dataDefs){
			Type dataType = new ADT(dataDef.name,dataDef.vars.stream().map((x)->new VarType(x)).collect(Collectors.toList()));
			for (ConDef conDef:dataDef.cons){
				conDef.type = dataType;
				cons.put(conDef.name, conDef);
			}
		}				
		
		for (FunDef fun: prog.funs){
			System.out.println(fun.name);
			fun.welcome(this);
		}

		return null;
	}

	@Override
	public Type visit(FunDef fun) throws Exception {
		env = new HashMap<String, Type>();
		
		for (VarDef arg:fun.args){
			env.put(arg.name, arg.type);
		}
		Type result = fun.def.welcome(this);

		findSub(fun.name,fun.type,result);
/*		if (!result.equals(fun.type)){
			throw new Exception("wrong result type for "+fun.name+ " "+result+" <-> "+fun.type);
		}*/
		return fun.type;
	}
	
	@Override
	public Type visit(Var v) throws Exception {
		v.type = env.get(v.name);
		if (v.type==null){
			FunDef fun = funs.get(v.name); 
			if (fun==null)
				throw new Exception("unknown variable: "+v.name);
			else {
				v.type = new FunType(fun.args.stream().map((x)->x.type).collect(Collectors.toList()), fun.type).fresh();
				return v.type;
			}
		}else{
			return v.type;		
		}
	}
	
	
	@Override
	public Type visit(LetExpr lambda) throws Exception {

		for (Pair<String, Tree> p:lambda.args){
			p.e2.welcome(this);
			env.put(p.e1, p.e2.type);
		}
		lambda.expr.welcome(this);
		lambda.type = lambda.expr.type;
		return lambda.type;
	}

	@Override
	public Type visit(CaseExpr aCase) throws Exception {
		throw new UnsupportedOperationException();
	}
	
	public Type visit(CaseExpr aCase,Type matchType) throws Exception {
		if (aCase.isVarCase){
			env.put(aCase.name, matchType);	
			return aCase.def.welcome(this);
			
		}
		ConDef con = cons.get(aCase.name);
		if (con==null)
			throw new Exception("unknown constructor name: "+aCase.name);
		Substitution freshSub = con.type.freshSub();
		Substitution sub = findSub("",con.type.substitute(freshSub),matchType);

		Iterator<VarDef> frmArgs = con.args.iterator();
		
		for (Var arg:aCase.args){
			Type type = frmArgs.next().type.substitute(freshSub).substitute(sub);
			env.put(arg.name, type);
		}
		Type result = aCase.def.welcome(this);
		return result;
	}

	@Override
	public Type visit(MatchExpr match) throws Exception {
		Type t = match.expr.welcome(this);
		if (!(t instanceof ADT))
			throw new Exception("match on non-algebraic type: "+t);
		for (CaseExpr aCase:match.cases){
			Map<String, Type> oldEnv = env;
			env = new HashMap<String, Type>();
			env.putAll(oldEnv);
			Type resultType = visit(aCase,t);
			env=oldEnv;
			if (match.type!=null){
				System.out.println("match "+match+"\n "+match.type+"\n "+resultType);
				Substitution mgu = findMGU("", match.type, resultType);
				match.type = match.type.substitute(mgu);
			}else {
				match.type = resultType;
			}
		}
		return match.type;
	}

	@Override
	public Type visit(ProjectionExpr lambda) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Type visit(LambdaExpr lambda) throws Exception {
		throw new UnsupportedOperationException();
	}


	@Override
	public Type visit(FunCall funcall) throws Exception {
		for (Tree arg:funcall.args){
			arg.welcome(this);
		}
		FunDef fun = funs.get(funcall.name);
		if (fun==null){			
			ConDef con = cons.get(funcall.name);
			if (con == null){
				Type var = env.get(funcall.name);
				if (var==null)
					throw new Exception("unknown function: "+funcall.name);
				while (var instanceof ADT){
					ADT adt = (ADT)var;
					TypeDef td = typs.get(adt.name);
					if (td!=null){
						var = td.instantiate(adt.args);
					}
				}
				if(!(var instanceof FunType)){
					throw new Exception("var "+funcall.name+" is not a function");					
				}
				List<Type> args = new ArrayList<Type>();
				FunType funType = (FunType)var;
				args.addAll(funType.args);
				args.add(funType.result);
				
				Substitution sub = findSub(funcall.name, funType.args, funcall.args.stream().map((x)->x.type).collect(Collectors.toList()));
				funcall.type = funType.result.substitute(sub);
				System.out.println(funcall.name+": "+funcall.type+" "+sub);
				return funcall.type;

				
			}else {
				Substitution freshSub = con.type.freshSub();
				Substitution sub =
				  findSub(funcall.name
					,     con.args.stream().map((x)->x.type.substitute(freshSub)).collect(Collectors.toList())
					, funcall.args.stream().map((x)->x.type).collect(Collectors.toList()),true);
				funcall.type = con.type.substitute(freshSub).substitute(sub);
				System.out.println(con.name+" "+freshSub+" "+sub+" "+funcall.type+" "+con.args+" "+funcall.args);
				return funcall.type;
			}
		}
		List<Type> args = new ArrayList<Type>();
		fun.args.forEach((x)->args.add(x.type));
		args.add(fun.type);
		Substitution freshSub = new ADT("Function",args).freshSub();
		
		List<Type> argsFRM = fun.args.stream().map((x)->x.type.substitute(freshSub)).collect(Collectors.toList());
		List<Type> argsACT = funcall.args.stream().map((x)->x.type).collect(Collectors.toList());
		System.out.println("substituiere für"+" "+funcall.name+"\n "+argsFRM+"\n "+argsACT);
		Substitution sub = findSub(funcall.name, argsFRM, argsACT,true);
		System.out.println("Sub gefunden: "+sub);
		funcall.type = fun.type.substitute(freshSub).substitute(sub);
		System.out.println(fun.name+": "+funcall.type+" "+freshSub+" "+sub);
		return funcall.type;
	}



	@Override
	public Type visit(ConDef fun) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Type visit(IntLit il) throws Exception {
		il.type = INT_TYPE;
		return il.type;
	}

	@Override
	public Type visit(StringLit il) throws Exception {
		List<Type> args = new ArrayList<>();
		args.add(CHAR_TYPE);
		il.type = new ADT("List", args);
		return il.type;
	}

	@Override
	public Type visit(MultExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = INT_TYPE;
		return ex.type;
	}

	@Override
	public Type visit(DivExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = INT_TYPE;
		return ex.type;
	}

	@Override
	public Type visit(ModExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = INT_TYPE;
		return ex.type;
	}

	private void visit(OpExpr ex) throws Exception {
		ex.left.welcome(this);
		ex.right.welcome(this);
	}

	@Override
	public Type visit(AddExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = INT_TYPE;
		return ex.type;
	}

	@Override
	public Type visit(SubExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = INT_TYPE;
		return ex.type;
	}

	boolean isPrimitive(Type type){
		return type.equals(BOOLEAN_TYPE)
				|| type.equals(INT_TYPE)
				|| type.equals(CHAR_TYPE);
	}
	
	@Override
	public Type visit(EqExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!isPrimitive(ex.left.type)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!isPrimitive(ex.right.type)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(NeqExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!isPrimitive(ex.left.type)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!isPrimitive(ex.right.type)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(LeExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(GeExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(LtExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(GtExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.left.type);
		};
		if (!ex.right.type.equals(INT_TYPE)){
			throw new Exception("Operand not numeric but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(OrExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(BOOLEAN_TYPE)){
			throw new Exception("Operand not boolean but: "+ex.left.type);
		};
		if (!ex.right.type.equals(BOOLEAN_TYPE)){
			throw new Exception("Operand not boolean but: "+ex.right.type);
		};
	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(AndExpr ex) throws Exception {
	    visit((OpExpr) ex);
		if (!ex.left.type.equals(BOOLEAN_TYPE)){
			throw new Exception("Operand not boolean but: "+ex.left.type);
		};
		if (!ex.right.type.equals(BOOLEAN_TYPE)){
			throw new Exception("Operand not boolean but: "+ex.right.type);
		};

	    ex.type = BOOLEAN_TYPE;
		return BOOLEAN_TYPE;
	}

	@Override
	public Type visit(IfExpr ex) throws Exception {
		ex.cond.welcome(this);
		System.out.println(ex.cond.type.getClass()+" "+BOOLEAN_TYPE.getClass()+" "+BOOLEAN_TYPE.name+" "+((PrimType)ex.cond.type).name);
		if (!ex.cond.type.equals(BOOLEAN_TYPE)){
			throw new Exception("condition in if is not of type boolean but: "+ex.cond.type);
		};
		
		ex.alt1.welcome(this);
		ex.alt2.welcome(this);
		System.out.println("TRY uification: "+ex.alt1.type+" "+ex.alt2.type);
		Substitution mgu = findMGU("", ex.alt1.type, ex.alt2.type);
		System.out.println(ex.alt1.type+" "+ex.alt2.type+" "+mgu);
//		if (!ex.alt1.type.equals(ex.alt2.type)){
		//	findSub("", formal, actual)
//			throw new Exception("different types of alternatives in if: "+ex.alt1.type+" "+ex.alt2.type);
//		};
		
		ex.type = ex.alt2.type.substitute(mgu);
		return ex.type;
	}

	@Override
	public Type visit(DataDef dataDef) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Type visit(VarDef varDef) throws Exception {
		throw new UnsupportedOperationException();
	}

	@Override
	public Type visit(BoolLit boolLit) throws Exception {
		boolLit.type = BOOLEAN_TYPE;
		return boolLit.type;
	}

}
